/*
 * Copyright (C) 2001-2002 Peter Weston 
 * peter.weston@holistech.co.uk
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 see end of this file for the full GNU General Public License
 
 MT: windows NT/2000/XP command line tape manipulation
 written to format a tape windows-XP backup refused to write or format... and satisfy my curiosity
 Can be used for crude data recovery (combine seek and read)

 todo: nothing... worked for me.
 coulddo: lots if anyone's interested...
		start read at n bytes...
		recover/retry errors (maybe reread damaged tape, pad unreadable block, seek over, etc.)
		gui
		write to tape
		fix long int nastyness
		test with other hardware
	 
  build with mingw (V3.2)... no makefile just a: gcc -o mt.exe mt.c
  
changelog:
v0.12		added eject as an alias for unload
v0.13   fixed broken seek number parsing
v0.14   error message tweaks
v0.15   fix for wrong block size during read (from Jason Bingham)
v0.16   add set block size and read variable block lengths
				
*/
#include <stdio.h>
#include <windows.h>

#define VER "v0.16"
#define MAX_ERROR_LINE_LENGTH 130
#define VARIABLE_BUFFER_LENGTH 131073  // 128K +1

enum {
	DO_ENDLIST,
	DO_LONGERASE,
	DO_REWIND,
	DO_DRIVESTATUS,
	DO_MEDIASTATUS,
	DO_FORMAT,
	DO_LOAD,
	DO_LOCK,
	DO_TENSION,
	DO_UNLOAD,
	DO_UNLOCK,
	DO_SHORTERASE,
	DO_USAGE,
	DO_PARTITION,
	DO_FIXEDPARTITION,
	DO_INITPARTITION,
	DO_SEEK,
	DO_READ,
	DO_LICENCE,
	DO_READPART,
	DO_BYTESEEK,
	DO_BLOCK,
	DO_EJECT,
} DO_ACTIONS;


typedef struct _PARRAY {
	char *txt;
	DWORD bit;
} PARRAY;

typedef struct _OPTARRAY {
	char *txt;
	DWORD bit;
	int params;
} OPTARRAY;


char licencetext[]={"
Copyright (C) 2001-2002 Peter Weston (peter.weston@holistech.co.uk)

this program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

this program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"};

char usagetext[]={"
Copyright (C) 2001-2002 Peter Weston (peter.weston@holistech.co.uk)
Usage: mt command <drive> <param1> <param2> ...
eg. mt erase tape0
Command, one of:
  block\t\t set block size to param1 bytes (eg. mt block 32K)
  drivestatus\t displays the status of the tape drive
  eject\t\t unlock then eject the tape
  format\t format the tape (QIC117 only)
  help\t\t display this usage information
  licence\t display licence information
  load\t\t load the tape
  lock\t\t stop tape being ejected
  longerase\t long-erase the specified tape (blanks current partition)
  mediastatus\t display the status of the media
  partition\t (no parameters) create fixed (drive default) partitions
  partition\t (1 parameter) create param1 select partitions
  partition\t (2 parameters) create param1 partitions of size param2 (MB)
  read\t\t (1 parameter) read from tape into file param1
  read\t\t (2 parameters) read param 2 bytes from tape to file param1
  retension\t retension the tape
  rewind\t rewind the tape
  seek\t\t move tape to block param1
  seekbyte\t move tape to byte param1
  shorterase\t write an end of tape marker at the current tape position
  unload\t unload the tape (without unlocking first - can deadlock/hang)
  unlock\t allow tape to be ejected

Drive:
  tape0 (default) tape1 etc...

Most commands fail while the tape is busy, except mediastatus which waits for
the current operation to complete.

Large numbers can be made by suffixing K (1024), M (K*1024) or G (M*1024)
"};

OPTARRAY opts[]={
	"longerase",DO_LONGERASE,0,
	"rewind",DO_REWIND,0,
	"drivestatus", DO_DRIVESTATUS,0,
	"mediastatus", DO_MEDIASTATUS,0,
	"format", DO_FORMAT,0,
	"load", DO_LOAD,0,
	"lock", DO_LOCK,0,
	"tension", DO_TENSION,0,
	"unload", DO_UNLOAD,0,
	"eject", DO_EJECT,0,
	"unlock", DO_UNLOCK,0,
	"shorterase", DO_SHORTERASE,0,
	"help", DO_USAGE,0,
	"partition", DO_FIXEDPARTITION,0,
	"partition", DO_PARTITION,1,
	"partition", DO_INITPARTITION,2,
	"seek", DO_SEEK,1,
	"seekbyte", DO_BYTESEEK,1,
	"read", DO_READ,1,
	"licence",DO_LICENCE,0,
	"read", DO_READPART,2,
	"block", DO_BLOCK,1,
	"",DO_ENDLIST,0
};

PARRAY drivel[]={
	"Hardware compression", TAPE_DRIVE_COMPRESSION,
	"Report if cleaning required",TAPE_DRIVE_CLEAN_REQUESTS,
	"Hardware error correction (ECC)",TAPE_DRIVE_ECC,
	"Soft eject physically ejects",TAPE_DRIVE_EJECT_MEDIA,
	"Must erase from partition start",TAPE_DRIVE_ERASE_BOP_ONLY,
	"Long erase\t\t",TAPE_DRIVE_ERASE_BOP_ONLY,
	"Now erase (immediate erase)",TAPE_DRIVE_ERASE_IMMEDIATE,
	"Short erase\t\t",TAPE_DRIVE_ERASE_SHORT,
	"Creates fixed partitions",TAPE_DRIVE_FIXED,
	"Fixed length blocks\t",TAPE_DRIVE_FIXED_BLOCK,
	"Initiator defined partitions",TAPE_DRIVE_INITIATOR,
	"Data padding\t\t",TAPE_DRIVE_PADDING,
	"Provides absolute block number",TAPE_DRIVE_GET_ABSOLUTE_BLK,
	"Provides logical block number",TAPE_DRIVE_GET_LOGICAL_BLK,
	"Setmark reporting\t",TAPE_DRIVE_REPORT_SMKS,
	"Creates select data partitions",TAPE_DRIVE_SELECT,
	"Set end of medium warning size",TAPE_DRIVE_SET_EOT_WZ_SIZE,
	"Set compression only at partition start",TAPE_DRIVE_SET_CMP_BOP_ONLY,
	"Returns maximum capacity of tape",TAPE_DRIVE_TAPE_CAPACITY,
	"Returns remaining capacity of tape",TAPE_DRIVE_TAPE_REMAINING,
	"Variable length block mode\t",TAPE_DRIVE_VARIABLE_BLOCK,
	"Returns error if write protected",TAPE_DRIVE_WRITE_PROTECT,
	"",0
};

PARRAY driveh[]={
	"Immediate return on absolute seek",TAPE_DRIVE_ABS_BLK_IMMED,
	"Absolute seek\t\t\t",TAPE_DRIVE_ABSOLUTE_BLK,
	"Goto end of data\t\t",TAPE_DRIVE_END_OF_DATA,
	"Seek by filemark count\t\t",TAPE_DRIVE_FILEMARKS,
	"Load and unload\t\t\t",TAPE_DRIVE_LOAD_UNLOAD,
	"Immediate return on load and unload",TAPE_DRIVE_LOAD_UNLD_IMMED,
	"Lock tape in drive\t\t",TAPE_DRIVE_LOCK_UNLOCK,
	"Immediate return on lock and unlock",TAPE_DRIVE_LOCK_UNLK_IMMED,
	"Immediate return on logical block seek",TAPE_DRIVE_LOG_BLK_IMMED,
	"Logical block seek\t",TAPE_DRIVE_LOGICAL_BLK,
	"Relative seek\t\t",TAPE_DRIVE_RELATIVE_BLKS,
	"Backwards seek\t\t",TAPE_DRIVE_REVERSE_POSITION,
	"Immediate rewind\t",TAPE_DRIVE_REWIND_IMMEDIATE,
	"Seek to n consecutive filemarks",TAPE_DRIVE_SEQUENTIAL_FMKS,
	"Seek to n consecutive setmarks",TAPE_DRIVE_SEQUENTIAL_SMKS,
	"Set block size\t\t",TAPE_DRIVE_SET_BLOCK_SIZE,
	"Set hardware compression",TAPE_DRIVE_SET_COMPRESSION,
	"Set hardware error correction",TAPE_DRIVE_SET_ECC,
	"Set data padding\t",TAPE_DRIVE_SET_PADDING,
	"Setmark reporting\t",TAPE_DRIVE_SET_REPORT_SMKS,
	"Seek over n setmarks\t",TAPE_DRIVE_SETMARKS,
	"Immediate spacing\t",TAPE_DRIVE_SPACE_IMMEDIATE,
	"Tape tensioning\t\t",TAPE_DRIVE_TENSION,
	"Immediate tape tensioning",TAPE_DRIVE_TENSION_IMMED,
	"Write filemarks\t\t",TAPE_DRIVE_WRITE_FILEMARKS,
	"Write long filemarks\t",TAPE_DRIVE_WRITE_LONG_FMKS,
	"Immediate write filemarks",TAPE_DRIVE_WRITE_MARK_IMMED,
	"Write setmarks\t\t",TAPE_DRIVE_WRITE_SETMARKS,
	"Write short setmarks\t",TAPE_DRIVE_WRITE_SHORT_FMKS,
	"",0
};

static char *TXT_PARTIONED_OK = "Partitioned OK\n";
static char *TXT_UNLOCKING = "unlocking";
static char *TXT_UNLOADING = "unloading";

void usage(void) {
  printf("\nMT: command-line tape manipulation similar to the unix mt command %s\n%s",VER,usagetext);
  exit(1);
}

void error(char* txt) {
	LPVOID buff;
	DWORD lasterr;
	
	lasterr = GetLastError();
	FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS + MAX_ERROR_LINE_LENGTH,
    NULL, lasterr, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR) &buff,0,NULL );
	if(buff!=NULL) {
		printf("%s While %s (%d)\n", (LPCTSTR)buff, txt ,lasterr);
		LocalFree(buff);
	} else {
		printf("unknown error %d while %s\n", lasterr, txt);
	}	
}

void printparam(char* txt, BOOLEAN yn) {
	printf("%s: %s\n",txt,(yn) ? "Yes" : "No");
}

void prnfeat(char* txt, DWORD result) {
	printf("%s\t: %s\n",txt,(result!=0) ? "Yes" : "No");
}

void hexdec(char *txt, DWORD val) {
	printf("%s: 0x%.8x (%u)\n",txt,val,val);
}

void hexdeclargeint(char *txt, LARGE_INTEGER val) {
	printf("%s: 0x%.8I64x (%I64u)\n",txt,val,val);
}

void printmediaparams(TAPE_GET_MEDIA_PARAMETERS mediaparams) {
	hexdeclargeint("Capacity (Bytes)\t",mediaparams.Capacity);
	hexdeclargeint("Remaining (Bytes)\t",mediaparams.Remaining);
	hexdec("Block Size\t\t",mediaparams.BlockSize);
	hexdec("Partition Count\t\t",mediaparams.PartitionCount);
	printparam("Write Protected\t\t",mediaparams.WriteProtected);
}

void printdriveparams(TAPE_GET_DRIVE_PARAMETERS driveparams) {
	int count;
	
	printparam("Hardware ECC\t\t",driveparams.ECC);
	printparam("Compression\t\t",driveparams.Compression);
	printparam("Data Padding\t\t",driveparams.DataPadding);
	printparam("Report Setmarks\t\t",driveparams.ReportSetmarks);
	hexdec("Default Block Size\t",driveparams.DefaultBlockSize);
	hexdec("Maximum Block Size\t",driveparams.MaximumBlockSize);
	hexdec("Minimum Block Size\t",driveparams.MinimumBlockSize);
	hexdec("Maximum Partitions\t",driveparams.MaximumPartitionCount);
	hexdec("EOT Warning Zone Size\t",driveparams.EOTWarningZoneSize);
	for(count=0;drivel[count].bit!=0;count++) {
		prnfeat(drivel[count].txt,drivel[count].bit & driveparams.FeaturesLow);
	}
	for(count=0;driveh[count].bit!=0;count++) {
		prnfeat(driveh[count].txt,driveh[count].bit & driveparams.FeaturesHigh);
	}
}

void closeh(HANDLE tapedrive) {
	if (CloseHandle(tapedrive)==FALSE) error("closing tape handle");
}

void errorclose(char *txt, HANDLE tapedrive) {
	error(txt);
	closeh(tapedrive);
	exit(1);
}

void preptape(HANDLE tapedrive, DWORD cmd, char *txt) {
	if (PrepareTape(tapedrive,cmd,FALSE)!=NO_ERROR) {
		errorclose(txt,tapedrive);
	} else {
		printf("%s completed OK\n", txt);
	}
}

void doerase(HANDLE tapedrive, DWORD cmd, char *txt) {
	if (EraseTape(tapedrive,cmd,TRUE)!=NO_ERROR) {
		errorclose(txt,tapedrive);
	} else {
		printf("%s completed OK\n",txt);
	}
}

DWORD64 GetNumber(char *txt) {
  char ltxt[101];  // v0.16 stopped corrupting passed string...
	unsigned char unit=0; // number of bits to shift result
	DWORD64 number=0;
	unsigned char c;
	for (c=0;(c<100) && (txt[c]!='\0');c++) {
		switch (txt[c]) {
			case 'k':
			case 'K':
				unit=10; // 1Kb in bits to shift
				break;
			case 'm':
			case 'M':
				unit=20;  // 1Mb in bits to shift
				break;
			case 'G':
			case 'g':
				unit=30; // 1Gb in bits to shift
				break;
		}
		if ((txt[c]<'0') || (txt[c]>'9')) break;
		ltxt[c]=txt[c];
	}
	ltxt[c]='\0';
	sscanf(ltxt,"%I64d",&number);
	return number<<unit;
}

#define getdword(a) ((DWORD) (GetNumber(a) & 0xffffffff))  // tisk

DWORD getdeviceblocksize(HANDLE tapedrive) {
	TAPE_GET_DRIVE_PARAMETERS driveparams;
	DWORD junk;
	
	junk =sizeof(TAPE_GET_DRIVE_PARAMETERS);
	if (GetTapeParameters(tapedrive,GET_TAPE_DRIVE_INFORMATION,&junk,&driveparams)!=NO_ERROR) {
		errorclose("getting tape drive information (block size)",tapedrive);
	} else {  // warn about block size under 1KB or over 1Mb
		if ((driveparams.DefaultBlockSize<1024) || (driveparams.DefaultBlockSize>1048576)) {
			printf("Warning: Drive reports block size as %d bytes\n",driveparams.DefaultBlockSize);
		}
		return driveparams.DefaultBlockSize;
	}
}

DWORD getmediablocksize(HANDLE tapedrive) {
	TAPE_GET_MEDIA_PARAMETERS mediaparams;
	DWORD junk;
	junk =sizeof(TAPE_GET_MEDIA_PARAMETERS);
	if (GetTapeParameters(tapedrive,GET_TAPE_MEDIA_INFORMATION,&junk,&mediaparams)!=NO_ERROR) {
		errorclose("getting tape drive information (block size)",tapedrive);
	} else {
		return mediaparams.BlockSize;
	}
}

void setmediablocksize(HANDLE tapedrive, DWORD blocks) {
	TAPE_SET_MEDIA_PARAMETERS mediaparams;

	mediaparams.BlockSize = blocks;
	if (SetTapeParameters(tapedrive,SET_TAPE_MEDIA_INFORMATION,&mediaparams)!=NO_ERROR) {
		printf("Use drivestatus to check the block size (%u) is supported by the drive\n",blocks);
		errorclose("setting block size",tapedrive);
	}
}

void printtapeposition(HANDLE tapedrive) {
	DWORD part, posl, posh;
	DWORD64 bytePosition;
	if (GetTapePosition(tapedrive,TAPE_ABSOLUTE_POSITION,&part, &posl, &posh)!=NO_ERROR) {
		errorclose("getting tape position",tapedrive);
	} else {
		printf("Current Partition\t: %u\n",part);
		printf("Absolute Block Position\t: 0x%.8x 0x%.8x (%d %d)\n",posh, posl,posh, posl);
		bytePosition=(((DWORD64)posh<<32)+posl)*getdeviceblocksize(tapedrive);
		printf("Absolute Byte Position\t: 0x%.16I64x\n",bytePosition);
	}
	if (GetTapeStatus(tapedrive)!=NO_ERROR) {
		errorclose("getting tape status", tapedrive);
	} else {
		printf("Status\t\t\t: Ready\n");
	}
}

BOOL variableblocks(HANDLE tapedrive) {
	TAPE_GET_DRIVE_PARAMETERS driveparams;
	DWORD junk;
	junk =sizeof(TAPE_GET_DRIVE_PARAMETERS);
	if (GetTapeParameters(tapedrive,GET_TAPE_DRIVE_INFORMATION,&junk,&driveparams)!=NO_ERROR) {
		errorclose("getting drive information (variable block support)",tapedrive);
	} else {
		return (driveparams.FeaturesLow & TAPE_DRIVE_VARIABLE_BLOCK);
	}
}

void readtape(HANDLE tapedrive, char *filename, char *length) {
	LPVOID buff;
	HANDLE outf;
	DWORD byteswritten;
	DWORD readbuffersize;
	DWORD64 toread;
	DWORD result=1;
	DWORD bytesread=1;
	DWORD blockcount=0;
	DWORD64 total=0;

	if (variableblocks(tapedrive)) {		// v0.16
		readbuffersize=VARIABLE_BUFFER_LENGTH;
		setmediablocksize(tapedrive,0);
	} else {
		readbuffersize=getmediablocksize(tapedrive);  // v0.15 media not device block size
	}
	if (length[0]=='\0') { 
		toread=-1;   // setting an unsigned to -1?
	} else {
		toread=GetNumber(length);
		if ((toread==0) && (GetNumber(filename)>0)) {  // 0.16 - no 0 byte files with numeric names
			printf("Warning: swapping filename ('%s') and bytes to read ('%s')\n",filename,length);
			toread=GetNumber(filename);
			filename=length;
		}
	}
	printtapeposition(tapedrive);
  buff=VirtualAlloc(NULL,readbuffersize,MEM_COMMIT,PAGE_READWRITE);
  if (buff==NULL) errorclose("allocating buffer",tapedrive);
	outf = CreateFile(filename,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
	if (outf==INVALID_HANDLE_VALUE) errorclose("opening output file",tapedrive);
	while ((bytesread>0) && (result!=0) && ((toread==-1) || (total<toread)))  {
		result = ReadFile(tapedrive,buff,readbuffersize,&bytesread,NULL);
		total+=bytesread;
		blockcount++;
		printf("Bytes read: %I64d from %u blocks\r",total,blockcount);
		if (WriteFile(outf,buff,bytesread,&byteswritten,NULL)==0) {
			closeh(outf);
			errorclose("writing file",tapedrive);
		}
	}
	closeh(outf);
	if (result==0) errorclose("reading tape",tapedrive);
	printf("%I64d bytes read from %u blocks \n",total,blockcount);
	if (toread!=total) printf("Rounded requested %I64d bytes up to the end of the last %d byte block\n",toread,bytesread);
	printtapeposition(tapedrive);
}

int main(int argc, char **argv) {
	char filename[512]="//./";
	HANDLE tapedrive;
	TAPE_GET_DRIVE_PARAMETERS driveparams;
	TAPE_GET_MEDIA_PARAMETERS mediaparameters;
	DWORD junk;
	int command=DO_ENDLIST;
	DWORD64 position;
	int count;
	int p1;
	
	if (argc<2) usage();
	if ((argc>2) && (strncasecmp(argv[2],"tape",4)==0)) {
		strncat(filename,argv[2],500);
		p1=3;
	} else {
		strncat(filename,"tape0",200);
		p1=2;
	}
	
	for (count=0; opts[count].bit!=0; count++) {
		if ((strcasecmp(argv[1],opts[count].txt)==0) && (argc==p1+opts[count].params)) {
			command=opts[count].bit;
			break;
		}
	}
	if ((command==DO_ENDLIST) || (command==DO_USAGE))  usage();
	tapedrive=CreateFile(filename,GENERIC_READ | GENERIC_WRITE,0,0,OPEN_EXISTING,0,NULL);
	if (tapedrive==INVALID_HANDLE_VALUE) {
		error("opening tape drive");
		exit(1);
	}
	switch(command) {
		case DO_LONGERASE:
			doerase(tapedrive,TAPE_ERASE_LONG,"long erase");
			break;
		case DO_SHORTERASE: 
			doerase(tapedrive,TAPE_ERASE_SHORT,"short erase");
			break;
		case DO_FORMAT:
			preptape(tapedrive,TAPE_FORMAT,"formatting. Try longerase for non QIC117 tape drives");
			break;
		case DO_LOAD:
			preptape(tapedrive,TAPE_LOAD,"loading");
			break;
		case DO_LOCK:
			preptape(tapedrive,TAPE_LOCK,"locking");
		  printf("warning: unlock tape before unloading\n");
			break;
		case DO_TENSION:
			preptape(tapedrive,TAPE_TENSION,"retensioning");
			break;
		case DO_UNLOAD:
		  printf("warning: unload may hang if the tape is locked\n");
			preptape(tapedrive,TAPE_UNLOAD,TXT_UNLOADING);
			break;
		case DO_UNLOCK:
			preptape(tapedrive,TAPE_UNLOCK,TXT_UNLOCKING);
			break;
		case DO_EJECT:
			preptape(tapedrive,TAPE_UNLOCK,TXT_UNLOCKING);
			preptape(tapedrive,TAPE_UNLOAD,TXT_UNLOADING);
			break;
		case DO_FIXEDPARTITION:
			if (CreateTapePartition(tapedrive,TAPE_FIXED_PARTITIONS,0,0)!=NO_ERROR) {
				errorclose("fixed partitioning",tapedrive);
			} else {
				printf(TXT_PARTIONED_OK);
			}
			break;
		case DO_INITPARTITION:
			if (CreateTapePartition(tapedrive,TAPE_INITIATOR_PARTITIONS,getdword(argv[p1]),getdword(argv[p1+1]))!=NO_ERROR) {
				errorclose("initiator partitioning",tapedrive);
			} else {
				printf(TXT_PARTIONED_OK);
			}
			break;
		case DO_PARTITION:
			if (CreateTapePartition(tapedrive,TAPE_SELECT_PARTITIONS,getdword(argv[p1]),0)!=NO_ERROR) {
				errorclose("select partitioning",tapedrive);
			} else {
				printf(TXT_PARTIONED_OK);
			}
			break;
		case DO_REWIND: 
			if (SetTapePosition(tapedrive,TAPE_REWIND,0,0,0,TRUE)!=NO_ERROR) {
				errorclose("rewinding",tapedrive);
			} else {
				printtapeposition(tapedrive);
			}
			break;
		case DO_DRIVESTATUS:
			junk =sizeof(TAPE_GET_DRIVE_PARAMETERS);
			if (GetTapeParameters(tapedrive,GET_TAPE_DRIVE_INFORMATION,&junk,&driveparams)!=NO_ERROR) {
				errorclose("getting tape drive information",tapedrive);
			} else {
				printdriveparams(driveparams);
			}
			break;
		case DO_MEDIASTATUS:
			junk =sizeof(TAPE_GET_MEDIA_PARAMETERS);
			if (GetTapeParameters(tapedrive,GET_TAPE_MEDIA_INFORMATION,&junk,&mediaparameters)!=NO_ERROR) {
				errorclose("getting media information",tapedrive);
			} else {
				printmediaparams(mediaparameters);
				printtapeposition(tapedrive);
			}
			break;
		case DO_SEEK:
		case DO_BYTESEEK:
			position = GetNumber(argv[p1]);
			if (command==DO_BYTESEEK) position=position/getdeviceblocksize(tapedrive);
			if (SetTapePosition(tapedrive,TAPE_ABSOLUTE_BLOCK,0,(DWORD) (position & 0xffffffff),(DWORD) ((position>>32) & 0xffffffff),FALSE)!=NO_ERROR) {
				errorclose("setting position",tapedrive);
			} else {
				printtapeposition(tapedrive);
			}
			break;
		case DO_READ:
			readtape(tapedrive,argv[p1],"");
			break;
		case DO_LICENCE:
			printf("%s",licencetext);
			break;
		case DO_READPART:
			readtape(tapedrive,argv[p1],argv[p1+1]);
			break;
		case DO_BLOCK:
			setmediablocksize(tapedrive,GetNumber(argv[p1]));
			break;
		default:
			usage();
	}
	closeh(tapedrive);
	return 0;
}

/* 
		    GNU GENERAL PUBLIC LICENSE
		       Version 2, June 1991

 Copyright (C) 1989, 1991 Free Software Foundation, Inc.
                       59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 Everyone is permitted to copy and distribute verbatim copies
 of this license document, but changing it is not allowed.

			    Preamble

  The licenses for most software are designed to take away your
freedom to share and change it.  By contrast, the GNU General Public
License is intended to guarantee your freedom to share and change free
software--to make sure the software is free for all its users.  This
General Public License applies to most of the Free Software
Foundation's software and to any other program whose authors commit to
using it.  (Some other Free Software Foundation software is covered by
the GNU Library General Public License instead.)  You can apply it to
your programs, too.

  When we speak of free software, we are referring to freedom, not
price.  Our General Public Licenses are designed to make sure that you
have the freedom to distribute copies of free software (and charge for
this service if you wish), that you receive source code or can get it
if you want it, that you can change the software or use pieces of it
in new free programs; and that you know you can do these things.

  To protect your rights, we need to make restrictions that forbid
anyone to deny you these rights or to ask you to surrender the rights.
These restrictions translate to certain responsibilities for you if you
distribute copies of the software, or if you modify it.

  For example, if you distribute copies of such a program, whether
gratis or for a fee, you must give the recipients all the rights that
you have.  You must make sure that they, too, receive or can get the
source code.  And you must show them these terms so they know their
rights.

  We protect your rights with two steps: (1) copyright the software, and
(2) offer you this license which gives you legal permission to copy,
distribute and/or modify the software.

  Also, for each author's protection and ours, we want to make certain
that everyone understands that there is no warranty for this free
software.  If the software is modified by someone else and passed on, we
want its recipients to know that what they have is not the original, so
that any problems introduced by others will not reflect on the original
authors' reputations.

  Finally, any free program is threatened constantly by software
patents.  We wish to avoid the danger that redistributors of a free
program will individually obtain patent licenses, in effect making the
program proprietary.  To prevent this, we have made it clear that any
patent must be licensed for everyone's free use or not licensed at all.

  The precise terms and conditions for copying, distribution and
modification follow.

		    GNU GENERAL PUBLIC LICENSE
   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION

  0. This License applies to any program or other work which contains
a notice placed by the copyright holder saying it may be distributed
under the terms of this General Public License.  The "Program", below,
refers to any such program or work, and a "work based on the Program"
means either the Program or any derivative work under copyright law:
that is to say, a work containing the Program or a portion of it,
either verbatim or with modifications and/or translated into another
language.  (Hereinafter, translation is included without limitation in
the term "modification".)  Each licensee is addressed as "you".

Activities other than copying, distribution and modification are not
covered by this License; they are outside its scope.  The act of
running the Program is not restricted, and the output from the Program
is covered only if its contents constitute a work based on the
Program (independent of having been made by running the Program).
Whether that is true depends on what the Program does.

  1. You may copy and distribute verbatim copies of the Program's
source code as you receive it, in any medium, provided that you
conspicuously and appropriately publish on each copy an appropriate
copyright notice and disclaimer of warranty; keep intact all the
notices that refer to this License and to the absence of any warranty;
and give any other recipients of the Program a copy of this License
along with the Program.

You may charge a fee for the physical act of transferring a copy, and
you may at your option offer warranty protection in exchange for a fee.

  2. You may modify your copy or copies of the Program or any portion
of it, thus forming a work based on the Program, and copy and
distribute such modifications or work under the terms of Section 1
above, provided that you also meet all of these conditions:

    a) You must cause the modified files to carry prominent notices
    stating that you changed the files and the date of any change.

    b) You must cause any work that you distribute or publish, that in
    whole or in part contains or is derived from the Program or any
    part thereof, to be licensed as a whole at no charge to all third
    parties under the terms of this License.

    c) If the modified program normally reads commands interactively
    when run, you must cause it, when started running for such
    interactive use in the most ordinary way, to print or display an
    announcement including an appropriate copyright notice and a
    notice that there is no warranty (or else, saying that you provide
    a warranty) and that users may redistribute the program under
    these conditions, and telling the user how to view a copy of this
    License.  (Exception: if the Program itself is interactive but
    does not normally print such an announcement, your work based on
    the Program is not required to print an announcement.)

These requirements apply to the modified work as a whole.  If
identifiable sections of that work are not derived from the Program,
and can be reasonably considered independent and separate works in
themselves, then this License, and its terms, do not apply to those
sections when you distribute them as separate works.  But when you
distribute the same sections as part of a whole which is a work based
on the Program, the distribution of the whole must be on the terms of
this License, whose permissions for other licensees extend to the
entire whole, and thus to each and every part regardless of who wrote it.

Thus, it is not the intent of this section to claim rights or contest
your rights to work written entirely by you; rather, the intent is to
exercise the right to control the distribution of derivative or
collective works based on the Program.

In addition, mere aggregation of another work not based on the Program
with the Program (or with a work based on the Program) on a volume of
a storage or distribution medium does not bring the other work under
the scope of this License.

  3. You may copy and distribute the Program (or a work based on it,
under Section 2) in object code or executable form under the terms of
Sections 1 and 2 above provided that you also do one of the following:

    a) Accompany it with the complete corresponding machine-readable
    source code, which must be distributed under the terms of Sections
    1 and 2 above on a medium customarily used for software interchange; or,

    b) Accompany it with a written offer, valid for at least three
    years, to give any third party, for a charge no more than your
    cost of physically performing source distribution, a complete
    machine-readable copy of the corresponding source code, to be
    distributed under the terms of Sections 1 and 2 above on a medium
    customarily used for software interchange; or,

    c) Accompany it with the information you received as to the offer
    to distribute corresponding source code.  (This alternative is
    allowed only for noncommercial distribution and only if you
    received the program in object code or executable form with such
    an offer, in accord with Subsection b above.)

The source code for a work means the preferred form of the work for
making modifications to it.  For an executable work, complete source
code means all the source code for all modules it contains, plus any
associated interface definition files, plus the scripts used to
control compilation and installation of the executable.  However, as a
special exception, the source code distributed need not include
anything that is normally distributed (in either source or binary
form) with the major components (compiler, kernel, and so on) of the
operating system on which the executable runs, unless that component
itself accompanies the executable.

If distribution of executable or object code is made by offering
access to copy from a designated place, then offering equivalent
access to copy the source code from the same place counts as
distribution of the source code, even though third parties are not
compelled to copy the source along with the object code.

  4. You may not copy, modify, sublicense, or distribute the Program
except as expressly provided under this License.  Any attempt
otherwise to copy, modify, sublicense or distribute the Program is
void, and will automatically terminate your rights under this License.
However, parties who have received copies, or rights, from you under
this License will not have their licenses terminated so long as such
parties remain in full compliance.

  5. You are not required to accept this License, since you have not
signed it.  However, nothing else grants you permission to modify or
distribute the Program or its derivative works.  These actions are
prohibited by law if you do not accept this License.  Therefore, by
modifying or distributing the Program (or any work based on the
Program), you indicate your acceptance of this License to do so, and
all its terms and conditions for copying, distributing or modifying
the Program or works based on it.

  6. Each time you redistribute the Program (or any work based on the
Program), the recipient automatically receives a license from the
original licensor to copy, distribute or modify the Program subject to
these terms and conditions.  You may not impose any further
restrictions on the recipients' exercise of the rights granted herein.
You are not responsible for enforcing compliance by third parties to
this License.

  7. If, as a consequence of a court judgment or allegation of patent
infringement or for any other reason (not limited to patent issues),
conditions are imposed on you (whether by court order, agreement or
otherwise) that contradict the conditions of this License, they do not
excuse you from the conditions of this License.  If you cannot
distribute so as to satisfy simultaneously your obligations under this
License and any other pertinent obligations, then as a consequence you
may not distribute the Program at all.  For example, if a patent
license would not permit royalty-free redistribution of the Program by
all those who receive copies directly or indirectly through you, then
the only way you could satisfy both it and this License would be to
refrain entirely from distribution of the Program.

If any portion of this section is held invalid or unenforceable under
any particular circumstance, the balance of the section is intended to
apply and the section as a whole is intended to apply in other
circumstances.

It is not the purpose of this section to induce you to infringe any
patents or other property right claims or to contest validity of any
such claims; this section has the sole purpose of protecting the
integrity of the free software distribution system, which is
implemented by public license practices.  Many people have made
generous contributions to the wide range of software distributed
through that system in reliance on consistent application of that
system; it is up to the author/donor to decide if he or she is willing
to distribute software through any other system and a licensee cannot
impose that choice.

This section is intended to make thoroughly clear what is believed to
be a consequence of the rest of this License.

  8. If the distribution and/or use of the Program is restricted in
certain countries either by patents or by copyrighted interfaces, the
original copyright holder who places the Program under this License
may add an explicit geographical distribution limitation excluding
those countries, so that distribution is permitted only in or among
countries not thus excluded.  In such case, this License incorporates
the limitation as if written in the body of this License.

  9. The Free Software Foundation may publish revised and/or new versions
of the General Public License from time to time.  Such new versions will
be similar in spirit to the present version, but may differ in detail to
address new problems or concerns.

Each version is given a distinguishing version number.  If the Program
specifies a version number of this License which applies to it and "any
later version", you have the option of following the terms and conditions
either of that version or of any later version published by the Free
Software Foundation.  If the Program does not specify a version number of
this License, you may choose any version ever published by the Free Software
Foundation.

  10. If you wish to incorporate parts of the Program into other free
programs whose distribution conditions are different, write to the author
to ask for permission.  For software which is copyrighted by the Free
Software Foundation, write to the Free Software Foundation; we sometimes
make exceptions for this.  Our decision will be guided by the two goals
of preserving the free status of all derivatives of our free software and
of promoting the sharing and reuse of software generally.

			    NO WARRANTY

  11. BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED
OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS
TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE
PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,
REPAIR OR CORRECTION.

  12. IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,
INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING
OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED
TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY
YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER
PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGES.

*/
